'''
Created on Mar 19, 2021

@author: Juan L. Mateo
'''

import json
import re
from aceofbases import pam

class BaseEditor:
    '''
    Generic class to define the different base editors with their PAM types
    and the methods needed.
    The seed for Bowtie will be always 5 prime, as well as the window location,
    but with respect to the target sequence.
    '''
    def __init__(self, pam_str, window_start, window_end, optimal_window_start,
                 optimal_window_end, base_to_edit, target_size):
        self.pam = pam.factory(pam_str)
        self.window_start = window_start
        self.window_end = window_end
        self.optimal_window_start = optimal_window_start
        self.optimal_window_end = optimal_window_end
        self.base_to_edit = base_to_edit
        self.target_size = target_size

    def get_pam(self):
        "Returns the object representing the PAM"
        return self.pam

    def get_window_start(self):
        "Returns the initial position of the windown within the target sequence"
        return self.window_start

    def get_window_end(self):
        "Returns the final position of the windown within the target sequence"
        return self.window_end

    def get_optimal_window_start(self):
        "Returns the optimal initial position of the windown within the target sequence"
        return self.optimal_window_start

    def get_optimal_window_end(self):
        "Returns the optimal final position of the windown within the target sequence"
        return self.optimal_window_end

    def get_base_to_edit(self):
        "Returns the base to be edited"
        return self.base_to_edit

    def get_target_size(self):
        "Returns the target size"
        return self.target_size

    def score_target(self, target_seq):
        "Returns the score for the current target as the efficacy of the edit"
        return 0

class BE4_Gam(BaseEditor):
    "Class that encodes the BE4-Gam base editor"
    def __init__(self):
        super().__init__("NGG",2,10,3,8,'C',20)

    def score_target(self, target_seq):
        idx_cs = [m.start() for m in re.finditer("C",target_seq, re.I)]
        score = float('-inf')
        for idx in idx_cs:
            if idx<self.window_start:
                continue
            if idx>=self.window_end:
                break
            if self.optimal_window_start <= idx < self.optimal_window_end:
                current_score = 5
            else:
                current_score = 1
            if target_seq[idx-1] == "C":
                current_score -= 2
            if target_seq[idx-1] == "A":
                current_score -= 3
            if target_seq[idx-1] == "G":
                current_score -= 4
            if current_score>score:
                score=current_score
        return score

class ancBE4max(BaseEditor):
    "Class that encodes the ancBE4max base editor"
    def __init__(self):
        super().__init__("NGG",2,9,3,7,'C',20)

    def score_target(self, target_seq):
        idx_cs = [m.start() for m in re.finditer("C",target_seq, re.I)]
        score = float('-inf')
        for idx in idx_cs:
            if idx<self.window_start:
                continue
            if idx>=self.window_end:
                break
            if self.optimal_window_start <= idx < self.optimal_window_end:
                current_score = 10
            else:
                current_score = 2
            if target_seq[idx-1] == "C":
                current_score -= 2
            if target_seq[idx-1] == "A":
                current_score -= 2
            if target_seq[idx-1] == "G":
                current_score -= 8
            if current_score>score:
                score=current_score
        return score

class evoBE4max(BaseEditor):
    "Class that encodes the evoBE4max base editor"
    def __init__(self):
        super().__init__("NGG",0,11,3,8,'C',20)

    def score_target(self, target_seq):
        idx_cs = [m.start() for m in re.finditer("C",target_seq, re.I)]
        score = float('-inf')
        for idx in idx_cs:
            if idx<self.window_start:
                continue
            if idx>=self.window_end:
                break
            if self.optimal_window_start <= idx < self.optimal_window_end:
                current_score = 10
            else:
                current_score = 2
            if target_seq[idx-1] == "C":
                current_score -= 2
            if target_seq[idx-1] == "A":
                current_score -= 8
            if target_seq[idx-1] == "G":
                current_score -= 2
            if current_score>score:
                score=current_score
        return score

class ABE8e(BaseEditor):
    "class that encodes the ABE8e base editor"
    def __init__(self):
        super().__init__("NGG",2,9,3,8,'A',20)

    def score_target(self, target_seq):
        if "A" in target_seq[self.optimal_window_start:self.optimal_window_end]:
            return 10
        if "A" in target_seq[self.window_start:self.window_end]:
            return 2
        return 0

def factory(type):
    "Method following the factory patern to create the different base editor objects"
    editors = {
        "BE4-Gam": BE4_Gam,
        "ancBE4max": ancBE4max,
        "evoBE4max": evoBE4max,
        "ABE8e": ABE8e
        }
    if not type in editors:
        assert 0, "Wrong Base Editor type: " + type
    return editors[type]()

if __name__ == "__main__":
    dictionary = {}
    for cls in BaseEditor.__subclasses__():
        obj = cls()
        dictionary[obj.__class__.__name__] = obj.__dict__
    with open("base_editors.js", "w") as fp:
        fp.write("var jsonBaseEditors=")
        json.dump(dictionary, fp, default=vars)



'''
Created on Jul 24, 2018

@author: Juan L. Mateo
'''
from aceofbases.sequence_methods import reverse_complement

class OffTarget:
    "Class representing an off-taget site"
    # New offtarget site coming for forward search, TTTN
    def __new_fwd(self, chromosome, strand, start, substitutions, sequence, length_seq, length_pam, core_range):
        self.chromosome = chromosome
        self.strand = strand
        if strand == "+":  # the search is done with the forward sequence!
            self.sequence = list(sequence)  # to make the string modifiable
        else:
            self.sequence = list(reverse_complement(sequence))  # to make the string modifiable
        self.start = start  # assuming bed coordinates
        self.end = start + length_seq

        tmp = substitutions.split(",")

        ####self.mismatches = len(tmp) - 1
        self.mismatches = 0
        self.alignment = ['|'] * (length_seq - length_pam)
        self.score = 0
        for substitution in tmp:
            if not ":" in substitution: # No mismatch, ignoring
                continue
            [idx,base] = substitution.split(':')
            idx = int(idx)
            if strand == "+":
                self.sequence[idx] = base[0]
            else:
                self.sequence[idx] = reverse_complement(base[0])
            if idx < length_pam:  # The mismatch in the PAM is not considered for score calculation of alignment
                continue
            self.mismatches += 1
            self.score = self.score + pow(1.2, idx - length_pam + 1)
            self.alignment[idx - length_pam] = '-'
        if core_range != "NA" and core_range > 0:
            self.alignment = "PAM[" + "".join(self.alignment[:core_range]) + "]" + "".join(self.alignment[core_range:])
        else:
            self.alignment = "PAM" + "".join(self.alignment)
        self.sequence = "".join(self.sequence)
        # self.sequence = reverse_complement("".join(self.sequence))

    # New offtarget site coming for reverse search, NGG and other PAMs
    def __new_rev(self, chromosome, strand, start, substitutions, sequence, length_seq, length_pam, core_range):
        self.chromosome = chromosome
        if strand == "+":  # the search is done with the reverse complemented sequence!
            self.strand = "-"
            self.sequence = list(sequence)  # to make the string modifiable
        else:
            self.strand = "+"
            self.sequence = list(reverse_complement(sequence))  # to make the string modifiable
        self.start = start  # assuming bed coordinates
        self.end = start + length_seq

        tmp = substitutions.split(",")

        ####self.mismatches = len(tmp) - 1
        self.mismatches = 0
        self.alignment = ['|'] * (length_seq - length_pam)
        self.score = 0
        for substitution in tmp:
            if not ":" in substitution: # No mismatch, ignoring
                continue
            [idx, base] = substitution.split(':')
            idx = int(idx)
            if strand == "+":
                self.sequence[idx] = base[0]
            else:
                self.sequence[idx] = reverse_complement(base[0])
            if idx < length_pam:  # The mismatch in the PAM is not considered for score calculation of alignment
                continue
            self.mismatches += 1
            self.score = self.score + pow(1.2, length_seq - idx)
            self.alignment[length_seq - 1 - idx] = '-'
        if core_range != "NA" and core_range > 0:
            self.alignment = "".join(self.alignment[:-core_range]) + "[" + "".join(self.alignment[-core_range:]) + "]PAM"
        else:
            self.alignment = "".join(self.alignment) + "PAM"
        self.sequence = reverse_complement("".join(self.sequence))


    def __init__(self, forward, chromosome, strand, start, substitutions, sequence,
                 length_seq, length_pam, core_range):
        if forward:
            self.__new_fwd(chromosome, strand, start, substitutions, sequence, length_seq, length_pam, core_range)
        else:
            self.__new_rev(chromosome, strand, start, substitutions, sequence, length_seq, length_pam, core_range)

    def set_gene_info(self, exons, genes):
        "Sets the information relates to the closest gene"
        closest = exons.closest(self.chromosome, self.start, self.end)

        self.gene_id = closest[0]
        self.gene_name = closest[1]
        self.distance = closest[2]
        self.intragenic = genes.overlaps(self.chromosome, self.start, self.end)

    def get_genomic_coordinates(self):
        "Returns the off-target coordinates"
        return [self.chromosome, str(self.start + 1), str(self.end)]
    def get_bed_coordinates(self):
        "Return the off-target coordinates in bed format"
        return [self.chromosome, str(self.start), str(self.end)]



'''
Created on 16 Jun 2015

@author: juanlmateo
'''

from operator import attrgetter
from aceofbases.sequence_methods import get_window_translations

########################################################
# https://bsouthga.dev/posts/color-gradients-with-python
def hex_to_rgb(hex):
    ''' "#FFFFFF" -> [255,255,255] '''
    # Pass 16 to the integer function for change of base
    return [int(hex[i:i+2], 16) for i in range(1,6,2)]


def rgb_to_hex(rgb):
    ''' [255,255,255] -> "#FFFFFF" '''
    # Components need to be integers for hex to make sense
    rgb = [int(x) for x in rgb]
    return "#"+"".join(["0{0:x}".format(v) if v < 16 else
            "{0:x}".format(v) for v in rgb])

def color_dict(gradient):
    ''' Takes in a list of RGB sub-lists and returns dictionary of
    colors in RGB and hex form for use in a graphing function
    defined later on '''
    return {"hex":[rgb_to_hex(rgb) for rgb in gradient],
            "r":[rgb[0] for rgb in gradient],
            "g":[rgb[1] for rgb in gradient],
            "b":[rgb[2] for rgb in gradient]}


def linear_gradient(start_hex, end_hex="#FFFFFF", how_many=10):
    ''' returns a gradient list of (how_many) colors between
    two hex colors. start_hex and end_hex
    should be the full six-digit color string,
    inlcuding the number sign ("#FFFFFF") '''
    # Starting and ending colors in RGB form
    start = hex_to_rgb(start_hex)
    end = hex_to_rgb(end_hex)
    # Initilize a list of the output colors with the starting color
    rbg_list = [start]
    # Calcuate a color at each evenly spaced value of color from 1 to how_many
    for color in range(1, how_many):
    # Interpolate RGB vector for color at the current value of color
        curr_vector = [
            int(start[j] + (float(color)/(how_many-1))*(end[j]-start[j]))
            for j in range(3)
            ]
        # Add it to our list of output colors
        rbg_list.append(curr_vector)
    return color_dict(rbg_list)
########################################################

def get_coord_y(target_sites, width, box_height):
    '''Returns a list containing:
    a list with the value for the coordinate Y of each site the
    coordinate Y of the reference line (middle line to separate plus
    and minus strand) height of the box'''
    coord_y = [None] * len(target_sites)
    last_end_in_level = [-1] * len(target_sites)
    max_level_plus = 0
    max_level_minus = 0
    box_height = box_height + 2  # 2 px space between boxes
    for idx, target_site in enumerate(target_sites):
        # only sites in the forward strand
        if target_site.strand == "-":
            break
        for idx2 in range(len(target_sites)):
            if target_site.position > last_end_in_level[idx2]:
                coord_y[idx] = idx2 + 1
                last_end_in_level[idx2] = target_site.position + width
                max_level_plus = max(max_level_plus, coord_y[idx])
                break
    # now sites in reverse strand
    last_end_in_level = [-1] * len(target_sites)
    for idx, target_site in enumerate(target_sites):
        # only sites in the reverse strand
        if target_site.strand == "+":
            continue
        for idx2 in range(len(target_sites)):
            if target_site.position > last_end_in_level[idx2]:
                coord_y[idx] = idx2 + 1
                last_end_in_level[idx2] = target_site.position + width
                max_level_minus = max(max_level_minus, coord_y[idx])
                break

    middle_line = max_level_plus * box_height + 10
    height_box = (max_level_plus + max_level_minus) * box_height + 20
    # now adapt coordinates to the SVG system, (0,0) upper left corner
    for idx, target_site in enumerate(target_sites):
        if target_site.strand == "+":
            coord_y[idx] = (max_level_plus - coord_y[idx]) * box_height + 10
        else:
            coord_y[idx] = (coord_y[idx] + max_level_plus) * box_height + 12 - box_height
    return([coord_y, middle_line, height_box])

def scale_candidate_pos(position, coordinates_width, cds, base_editor, max_width=980.0, left_offset=10):
    "Returns the position of the candidates scaled wrt the plot length"
    sgrna_size = base_editor.get_target_size() + len(base_editor.pam.pam_str)
    real_position = position + sgrna_size + 1
    for idx in range(1,len(cds.exons)):
        exon1 = cds.exons[idx-1]
        exon2 = cds.exons[idx]
        if position < exon1.end-cds.start:
            break
        if position < exon2.start-cds.start:
            real_position -= (position - exon1.end + cds.start)
            real_position += sgrna_size - (exon2.start - cds.start - position)
            break
        real_position += sgrna_size - (exon2.start - exon1.end)

    return str(real_position * max_width / coordinates_width + left_offset)

def scale_x_axis_ticks(value_x, coordinates_width, cds, base_editor, max_width=980.0, left_offset=10):
    "Returns the scaled values of the X axis ticks"
    sgrna_size = base_editor.get_target_size() + len(base_editor.pam.pam_str)
    real_value_x = value_x + sgrna_size
    for exon in cds.exons:
        if value_x <= (exon.end-cds.start+1):
            break
        real_value_x += sgrna_size

    return str(real_value_x * max_width / coordinates_width + left_offset)

def scale_width(value_x, coordinates_width, max_width=980.0):
    "Returns the scaled value wrt the total width in the plot"
    return str(value_x * max_width / coordinates_width)

def get_axis_ticks(cds):
    "Returns a list with the values of the X axis ticks"
    tick_sep = 10
    first = 0
    ticks = list(range(first, cds.length, tick_sep))
    # a maximum of 10 ticks are returned
    return [1] + [ticks[m] for m in range(0,len(ticks),max(1,round(len(ticks)/10)))][1:]

def get_non_synonymous_changes(candidate, cds, query_length, base_editor, coord_y, box_height):
    "Returns the SVG code to represent non synonumous changes in the ORF"
    translations = get_window_translations(candidate,cds,base_editor)
    code_svg = ""
    scaled_pos = scale_candidate_pos(candidate.position, query_length, cds, base_editor)

    for frame,translation in enumerate(translations.frames):
        for idx, residue in enumerate(translation.translation):
            if ">" in residue:
                if "*" in residue:
                    color = "black"#"#d95f0e"
                else:
                    color = "purple"#"#fec44f"
                if candidate.strand == "+":
                    if frame<3:
                        nss_position = float(scaled_pos) + float(scale_width(idx * 3 - (translation.window_start_nt - base_editor.window_start), query_length))
                    else:
                        new_idx = len(translation.translation)-idx-1
                        new_window_start_nt = len(translation.translation)*3-translation.window_end_nt
                        nss_position = float(scaled_pos) + float(scale_width(new_idx * 3 - (new_window_start_nt - base_editor.window_start), query_length))
                else:
                    if frame<3:
                        rev_window_start_nt = len(candidate.sequence)-base_editor.window_end
                        nss_position = float(scaled_pos) + float(scale_width(idx*3 + rev_window_start_nt-translation.window_start_nt, query_length))
                    else:
                        new_idx = len(translation.translation)-idx-1
                        rev_window_start_nt = len(candidate.sequence)-base_editor.window_end
                        nss_position = float(scaled_pos) + float(scale_width(new_idx*3 + rev_window_start_nt-translation.window_end_nt, query_length))
                code_svg = code_svg \
                    + '<rect class="non-synonymous nss-F' + str((frame%3+1)*pow(-1,int(frame>2))) + '" style="display:none" x="' \
                    + str(nss_position) \
                    + '" y="' + str(coord_y) + '" fill="' + color \
                    + '" width="' + scale_width(3, query_length) + '" height="' + str(box_height) + '">' \
                    + '<title>' + residue + '</title></rect>\n'
    return code_svg

def generate_targets_pam_3prime(sorted_candidates, candidate_colors, cds, base_editor, coord_width, coord_y, box_height, font_size):
    "Returns the SVG code to represent the target sites when the PAM considered is 3 prime"
    pam_size = len(base_editor.pam.pam_str)
    target_size = base_editor.target_size
    code_svg = ""
    for idx, sorted_candidate in enumerate(sorted_candidates):
        scaled_pos = scale_candidate_pos(sorted_candidate.position,coord_width,cds,base_editor)
        scaled_pam_size = scale_width(pam_size, coord_width)
        scaled_target_size = scale_width(target_size, coord_width)
        scaled_size = str(float(scaled_pam_size)+float(scaled_target_size))
        scaled_text_pos = str(float(scaled_pos)-1)

        code_svg = code_svg + '<rect x="' + scaled_pos + '" y="' + str(coord_y[idx]) + '" opacity="0.5" fill="' + candidate_colors["hex"][int(sorted_candidate.label[1:])-1] + '" stroke="none" width="' + scaled_size + '" height="' + str(box_height) + '"/>\n'

        if sorted_candidate.strand == '+':
            code_svg = code_svg + '<rect x="' + str(float(scaled_pos)+float(scaled_target_size)) + '" y="' + str(coord_y[idx]) + '" fill="green" width="' + scaled_pam_size + '" height="' + str(box_height) + '"/><!-- PAM -->\n'\
            + '<a xlink:href="#' + sorted_candidate.label + '" title="' + sorted_candidate.label + '">\n'\
            + '<rect x="' + scaled_pos + '" y="' + str(coord_y[idx]) + '" opacity="0.1" fill="#FFFFFF" stroke="#000000" stroke-width="3" width="' + scaled_size + '" height="' + str(box_height) + '" onmouseover="fire(this);" onmouseout="setoff(this);"/>\n'
            # + '</a>\n'
        else:
            code_svg = code_svg + '<rect x="' + scaled_pos + '" y="' + str(coord_y[idx]) + '" fill="green" width="' + scaled_pam_size + '" height="' + str(box_height) + '"/><!-- PAM -->\n'\
            + '<a xlink:href="#' + sorted_candidate.label + '" title="' + sorted_candidate.label + '">\n'\
            + '<rect x="' + scaled_pos + '" y="' + str(coord_y[idx]) + '" opacity="0.1" fill="#FFFFFF" stroke="#000000" stroke-width="3" width="' + scaled_size + '" height="' + str(box_height) + '" onmouseover="fire(this);" onmouseout="setoff(this);"/>\n'
            # + '</a>\n'

        code_svg += get_non_synonymous_changes(sorted_candidate,cds,coord_width,base_editor,coord_y[idx],box_height)
        code_svg += '<text x="' + scaled_text_pos + '" y="' + str(coord_y[idx] + box_height - 3) + '" style="text-anchor: end" font-family="\'Helvetica\'" font-size="' + str(font_size) + '">' + sorted_candidate.label + '</text>\n'

        code_svg += '</a>\n'

    return code_svg

def get_svg(cds, candidates, base_editor):
    '''Returns the SVG code to represent the target sites and their editions
    that will affect the ORF.
    NOTE:
        cds is in gff format, i.e. 1-index
        target site position are in bed format, i.e. 0-index'''

    pam_size = len(base_editor.pam.pam_str)
    target_size = base_editor.get_target_size()
    code_svg = ""
    box_height = 18

    coord_width = cds.length + (target_size+pam_size) * 2
    if len(cds.exons)>1:
        coord_width += (target_size+pam_size) * (len(cds.exons)-1)

    if coord_width <= 500:
        font_size = 16
    elif coord_width <= 1000:
        font_size = 11
    elif coord_width <= 1500:
        font_size = 8
    else:
        font_size = 0

    # order by strand and position
    sorted_candidates = sorted(candidates.get_sites(), key=attrgetter('strand', 'position'))
    # each candidates needs horizontal space for the sequence plus the label
    width = target_size + pam_size + font_size
    [coord_y, middle_line, height_box] = get_coord_y(sorted_candidates, width, box_height)
    height_plot = height_box + 40

    axis_ticks = get_axis_ticks(cds)

    candidate_colors = linear_gradient("#fff7bc", "#d95f0e", len(sorted_candidates))

    code_svg = code_svg + '<svg version="1.1" id="Layer_1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" x="0px" y="0px" width="1000px" height="' + str(height_plot + 20) + 'px" viewBox="0 0 1000 ' + str(height_plot + 20) + '" enable-background="new 0 0 1000 ' + str(height_plot + 20) + '" xml:space="preserve">\n'
    code_svg = code_svg + '<rect fill="#DBDBDB" width="1000" height="' + str(height_box) + '"/>\n'
    code_svg = code_svg + '<line fill="none" stroke="#231F20" x1="10" y1="' + str(middle_line) + '" x2="990" y2="' + str(middle_line) + '"/>\n'

    for tick in axis_ticks:
        scaled_tick_pos = scale_x_axis_ticks(tick, coord_width, cds, base_editor)
        code_svg = code_svg + '<text x="' + scaled_tick_pos + '" y="' + str(height_box + 20) + '" style="text-anchor: middle" fill="#808080" font-family="\'Helvetica\'" font-size="12">' + str(tick) + '</text>\n'
        code_svg = code_svg + '<line fill="none" stroke="#EFEFEF" stroke-width="2" x1="' + scaled_tick_pos + '" y1="0" x2="' + scaled_tick_pos + '" y2="' + str(height_box) + '"/>\n'

    exon_border = 1
    for exon in cds.exons:
        pos = scale_x_axis_ticks(exon_border, coord_width, cds, base_editor)
        code_svg = code_svg + '<line stroke-dasharray="5,5" stroke="black" stroke-width="2" x1="' + pos + '" y1="0" x2="' + pos + '" y2="' + str(height_box) + '"/>\n'
        pos = scale_x_axis_ticks(exon_border + exon.end-exon.start, coord_width, cds, base_editor)
        code_svg = code_svg + '<line stroke-dasharray="5,5" stroke="black" stroke-width="2" x1="' + pos + '" y1="0" x2="' + pos + '" y2="' + str(height_box) + '"/>\n'
        exon_border +=exon.end-exon.start+1

    if base_editor.pam.is5prime:
        raise NotImplementedError("The SVG output not implemented yet for 5 prime PAMs")
    else:
        code_svg = code_svg + generate_targets_pam_3prime(sorted_candidates, candidate_colors, cds, base_editor, coord_width, coord_y, box_height, font_size)

    # code_svg = code_svg + '<text x="500" y="' + str(height_plot) + '" style="text-anchor: middle" font-family="\'Helvetica\'" font-size="18">' + cds.chrom + '</text>'
    code_svg = code_svg + '</svg>'

    return code_svg



'''
Created on Jul 20, 2018

@author: Juan L. Mateo
'''

import json
from aceofbases.sequence_methods import reverse_complement


class PAM:
    '''
    Generic class to define the different PAM types and the methods needed
    The seed for Bowtie will be always 5
    '''

    def __init__(self):
        '''
        Constructor
        '''
        self.core_allowed = False  # By default a PAM is not compatible with the core region
        self.is5prime = False # Is the PAM at the 5' end, by default it is false
        self.pam_str = self.__class__.__name__
        self.pam_rev = reverse_complement(self.pam_str)  # any ambiguous base is turned into N
        self.mismatches_in_pam = self.pam_rev.count('N')
        self.mismatches_in_seed = min(3,self.pam_rev[0:5].count('N')) # Value for the -n parameter of Bowtie


    def get_sequences_for_bowtie(self, sites):
        '''
        Receives a list of EditingSite objects and returns a string with the
        fasta sequences formatted to perform the search of off-target sites with
        Bowtie
        '''
        # By default all PAM are in the 3' end and the sequence is reversed to match the seed definition of Bowtie
        fasta = ""
        for candidate in sites:
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + self.pam_rev + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
        return fasta

    def get_bowtie_options(self, core_mismatches, core_range, total_mismatches):
        "Returns the options string to run bowtie"
        if not self.core_allowed or core_mismatches == "NA" or core_range == "NA":
            return ["-a", "--quiet", "-y", "-n" + str(self.mismatches_in_seed), "-l5",
                    "-e" + str((total_mismatches + self.mismatches_in_pam) * 30)]

        if (core_mismatches + self.mismatches_in_pam)>3:
            raise ValueError('The value for the parameter core_mismatches is not valid: ' + str(core_mismatches))
        if core_mismatches > total_mismatches:
            raise ValueError('The value for core_mismatches cannot be greater than total_mismatches:'
                             + str(core_mismatches) + ">" + str(total_mismatches))
        return ["-a", "--quiet", "-y", "-n" + str(core_mismatches + self.mismatches_in_pam),
                "-l" + str(core_range + len(self.pam_str)),
                "-e" + str((total_mismatches + self.mismatches_in_pam) * 30)]

    def is_bowtie_hit_consistent(self, bowtie_line_columns):
        '''
        Tells if the hit found show a set of mismatches, if any, consistent with the variable positions in the PAM
        '''
        # This generic version works for PAMs (5' or 3') where only the first position is an N
        # and the other position are not variables
        # NGG, NGA, NGCG, TTTN, NAAAAC
        # This is also fine for NRG because we search twice for NGG and NAG
        # The same for YTN (CTN, TTN)
        mismatches = bowtie_line_columns[7].split(',')
        pos_first_mm = int(mismatches[0].split(":")[0])
        return pos_first_mm >= (len(self.pam_str) - 1)

class NGG(PAM):
    "Class that represents the NGG PAM"
    def __init__(self):
        super().__init__()
        self.core_allowed = True
        self.mismatches_in_seed = 3

class NRG(PAM):
    "Class that represents the NRG PAM"
    def __init__(self):
        super().__init__()
        self.core_allowed = True
        self.mismatches_in_pam = 1
        self.mismatches_in_seed = 3

    def get_sequences_for_bowtie(self, sites):
        '''
        Receives a list of EditingSite objects and returns a string with the
        fasta sequences formatted to perform the search of off-target sites with
        Bowtie
        '''
        # As there are too many variable positions it is necessary to search
        # two sequences fixing the 'R' with 'A' or 'G' in each case
        fasta = ""
        for candidate in sites:
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "CCN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "CTN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
        return fasta

class NG(PAM):
    "Class that represents the NG PAM"
    def __init__(self):
        super().__init__()
        self.core_allowed = True
        self.mismatches_in_pam = 0
        self.mismatches_in_seed = 3

    def get_sequences_for_bowtie(self, sites):
        '''
        Receives a list of EditingSite objects and returns a string with the
        fasta sequences formatted to perform the search of off-target sites with
        Bowtie
        '''
        # As there are too many variable positions it is necessary to search
        # four sequences fixing the 'N' with the four possible bases
        fasta = ""
        for candidate in sites:
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "CT" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "CG" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "CC" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "CA" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
        return fasta
    def is_bowtie_hit_consistent(self, bowtie_line_columns):
        mismatches = bowtie_line_columns[7].split(',')
        # It is possible to get 0 mismatches, in that case it is consistent
        if not ":" in mismatches[0]:
            return True
        pos_first_mm = int(mismatches[0].split(":")[0])
        return pos_first_mm >= len(self.pam_str)

class NGA(PAM):
    "Class that represents the NGA PAM"
    def __init__(self):
        super().__init__()
        self.core_allowed = True
        self.mismatches_in_seed = 3

class NGCG(PAM):
    "Class that represents the NGCG PAM"
    def __init__(self):
        super().__init__()
        self.core_allowed = True
        self.mismatches_in_seed = 2

class TTTN(PAM):
    "Class that represents the TTTN PAM"
    def __init__(self):
        super().__init__()
        self.mismatches_in_seed = 2
        self.is5prime = True

    def get_sequences_for_bowtie(self, sites):
        '''
        Receives a list of EditingSite objects and returns a string with the
        fasta sequences formatted to perform the search of off-target sites with
        Bowtie
        '''
        # In this case the PAM is 5'
        fasta = ""
        pam_str = self.__class__.__name__
        for candidate in sites:
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + pam_str + candidate.sequence[len(pam_str):] + "\n"
        return fasta

class TTTV(PAM):
    "Class that represents the TTTV PAM"
    def __init__(self):
        super().__init__()
        self.mismatches_in_seed = 1
        self.is5prime = True
        self.mismatches_in_pam = 0 # The variable position is replaced by the 3 possible bases

    def get_sequences_for_bowtie(self, sites):
        '''
        Receives a list of EditingSite objects and returns a string with the
        fasta sequences formatted to perform the search of off-target sites with
        Bowtie
        '''
        # In this case the PAM is 5'
        fasta = ""
        pam_str = self.__class__.__name__
        for candidate in sites:
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "TTTA" + candidate.sequence[len(pam_str):] + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "TTTC" + candidate.sequence[len(pam_str):] + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "TTTG" + candidate.sequence[len(pam_str):] + "\n"
        return fasta

    def is_bowtie_hit_consistent(self, bowtie_line_columns):
        mismatches = bowtie_line_columns[7].split(',')
        # It is possible to get 0 mismatches, in that case it is consistent
        if not ":" in mismatches[0]:
            return True
        pos_first_mm = int(mismatches[0].split(":")[0])
        return pos_first_mm >= len(self.pam_str)

class YTN(PAM):
    "Class that represents the YTN PAM"
    def __init__(self):
        super().__init__()
        self.mismatches_in_pam = 1
        self.mismatches_in_seed = 3
        self.is5prime = True

    def get_sequences_for_bowtie(self, sites):
        '''
        Receives a list of EditingSite objects and returns a string with the
        fasta sequences formatted to perform the search of off-target sites with
        Bowtie
        '''
        # In this case the PAM is 5'
        fasta = ""
        pam_str = self.__class__.__name__
        for candidate in sites:
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "TTN" + candidate.sequence[len(pam_str):] + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "CTN" + candidate.sequence[len(pam_str):] + "\n"
        return fasta

class TTN(PAM):
    "Class that represents the TTN PAM"
    def __init__(self):
        super().__init__()
        self.mismatches_in_pam = 1
        self.mismatches_in_seed = 3
        self.is5prime = True

    def get_sequences_for_bowtie(self, sites):
        '''
        Receives a list of EditingSite objects and returns a string with the
        fasta sequences formatted to perform the search of off-target sites with
        Bowtie
        '''
        # In this case the PAM is 5'
        fasta = ""
        pam_str = self.__class__.__name__
        for candidate in sites:
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + pam_str + candidate.sequence[len(pam_str):] + "\n"
        return fasta

class YTTN(PAM):
    "Class that represents the YTTN PAM"
    def __init__(self):
        super().__init__()
        self.mismatches_in_pam = 1
        self.mismatches_in_seed = 2
        self.is5prime = True

    def get_sequences_for_bowtie(self, sites):
        # In this case the PAM is 5'
        fasta = ""
        pam_str = self.__class__.__name__
        for candidate in sites:
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "TTTN" + candidate.sequence[len(pam_str):] + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "CTTN" + candidate.sequence[len(pam_str):] + "\n"
        return fasta

class NNNRRT(PAM):
    "Class that represents the NNNRRT PAM"
    def __init__(self):
        super().__init__()
        # One of the variable position will be fixed so that the number or real mismatches is one less
        self.mismatches_in_pam = 4

    def get_sequences_for_bowtie(self, sites):
        '''
        Receives a list of EditingSite objects and returns a string with the
        fasta sequences formatted to perform the search of off-target sites with
        Bowtie
        '''
        # As there are too many variable positions it is necessary to search
        # two sequences fixing the most 3' 'R' with 'A' or 'G' in each case
        fasta = ""
        for candidate in sites:
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "ACNNNN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "ATNNNN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
        return fasta

    def is_bowtie_hit_consistent(self, bowtie_line_columns):
        mismatches = bowtie_line_columns[7].split(',')
        first_mm = mismatches[0].split(":")
        pos_first_mm = int(first_mm[0])
        if pos_first_mm > 2:
            return True
        if pos_first_mm == 2:
            if bowtie_line_columns[1] == '+':
                return first_mm[1][0] in 'CT'
            return first_mm[1][0] in 'AG'

        return False

class NNGRRT(PAM):
    "Class that represents the NNGRRT PAM"
    def is_bowtie_hit_consistent(self, bowtie_line_columns):
        mismatches = bowtie_line_columns[7].split(',')
        first_mm = mismatches[0].split(":")
        pos_first_mm = int(first_mm[0])
        if pos_first_mm > 3:
            return True
        if pos_first_mm == 1:
            return (bowtie_line_columns[1] == '+'
                    and (mismatches[0][2] in 'CT' and mismatches[1][2] in 'CT')) or (bowtie_line_columns[1] == '-' and (mismatches[0][2] in 'AG' and mismatches[1][2] in 'AG'))
        return False

class NNNNGATT(PAM):
    "Class that represents the NNNNGATT PAM"
    def is_bowtie_hit_consistent(self, bowtie_line_columns):
        return True

class NNAGAAW(PAM):
    "Class that represents the NNAGAAW PAM"
    def is_bowtie_hit_consistent(self, bowtie_line_columns):
        mismatches = bowtie_line_columns[7].split(',')
        first_mm = mismatches[0].split(":")
        pos_first_mm = int(first_mm[0])
        if pos_first_mm > 4:
            return True
        if pos_first_mm == 0:
            return mismatches[0][2] in 'AT'
        return False

class NAAAAC(PAM):
    "Class that represents the NAAAAC PAM"
    def is_bowtie_hit_consistent(self, bowtie_line_columns):
        return True

class NNNNRYAC(PAM):
    "Class that represents the NNNNRYAC PAM"
    def is_bowtie_hit_consistent(self, bowtie_line_columns):
        mismatches = bowtie_line_columns[7].split(',')
        first_mm = mismatches[0].split(":")
        pos_first_mm = int(first_mm[0])
        if pos_first_mm > 3:
            return True
        if pos_first_mm == 2:
            return (bowtie_line_columns[1] == '+' and (mismatches[0][2] in 'AG' and mismatches[1][2] in 'CT')) or (bowtie_line_columns[1] == '-' and (mismatches[0][2] in 'CT' and mismatches[1][2] in 'AG'))
        return False

class NRN(PAM):
    "Class that represents the NRN PAM"
    def __init__(self):
        super().__init__()
        self.mismatches_in_pam = 1

    def get_sequences_for_bowtie(self, sites):
        '''
        Receives a list of EditingSite objects and returns a string with the
        fasta sequences formatted to perform the search of off-target sites with
        Bowtie
        '''
        # By default all PAM are in the 3' end and the sequence is reversed to match the seed definition of Bowtie
        fasta = ""
        for candidate in sites:
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "ACN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "CCN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "GCN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "TCN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "ATN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "CTN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "GTN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "TTN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"

        return fasta

    def is_bowtie_hit_consistent(self, bowtie_line_columns):
        mismatches = bowtie_line_columns[7].split(',')
        pos_first_mm = int(mismatches[0].split(":")[0])
        if pos_first_mm > 1:
            return True
        return False

class NYN(PAM):
    "Class that represents the NYN PAM"
    def __init__(self):
        super().__init__()
        self.mismatches_in_pam = 1

    def get_sequences_for_bowtie(self, sites):
        '''
        Receives a list of EditingSite objects and returns a string with the
        fasta sequences formatted to perform the search of off-target sites with
        Bowtie
        '''
        # By default all PAM are in the 3' end and the sequence is reversed to match the seed definition of Bowtie
        fasta = ""
        for candidate in sites:
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "AAN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "CAN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "GAN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "TAN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "AGN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "CGN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "GGN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"
            fasta = fasta + ">" + candidate.label + "\n"
            fasta = fasta + "TGN" + reverse_complement(candidate.sequence[:-len(self.pam_str)]) + "\n"

        return fasta

    def is_bowtie_hit_consistent(self, bowtie_line_columns):
        '''
        Tells if the hit found show a set of mismatches, if any, consistent with the variable positions in the PAM
        '''
        mismatches = bowtie_line_columns[7].split(',')
        pos_first_mm = int(mismatches[0].split(":")[0])
        if pos_first_mm > 1:
            return True
        return False

def factory(type):
    "Method following the factory patern to create the different PAM objects"
    pams = {
        "NGG": NGG,
        "NRG": NRG,
        "NG": NG,
        "NGA": NGA,
        "NGCG": NGCG,
        "TTTN": TTTN,
        "TTTV": TTTV,
        "YTN": YTN,
        "TTN": TTN,
        "YTTN": YTTN,
        "NNNRRT": NNNRRT,
        "NNGRRT": NNGRRT,
        "NNNNGATT": NNNNGATT,
        "NNAGAAW": NNAGAAW,
        "NAAAAC": NAAAAC,
        "NNNNRYAC": NNNNRYAC,
        "NRN": NRN,
        "NYN": NYN,
        }
    if not type in pams:
        assert 0, "Wrong PAM type: " + type
    return pams[type]()



if __name__ == "__main__":
    dictionary = {}
    for cls in PAM.__subclasses__():
        obj = cls()
        dictionary[obj.__class__.__name__] = obj.__dict__
    with open("pams.js", "w") as fp:
        fp.write("var jsonPAMs=")
        json.dump(dictionary, fp, default=vars)



'''
Created on Mar 20, 2014

@author: juan
'''

import gzip

from bx.intervals.intersection import IntervalTree


class MyInterval():
    def __init__(self,start,end,value=None):
        self.start = start
        self.end = end
        self.value = value

class BedInterval():
    '''
    classdocs
    '''


    def __init__(self):
        '''
        Constructor
        '''
        self.chroms = {}

    def insert( self, chrom, start, end, gene_id=None, gene_name=None ):
        if chrom in self.chroms:
            if(gene_id is None or gene_name is None):
                self.chroms[chrom].insert( start, end, MyInterval(start,end) )
            else:
                self.chroms[chrom].insert( start, end, MyInterval(start,end,[gene_id, gene_name]) )
        else:
            self.chroms[chrom] = IntervalTree()
            if(gene_id is None or gene_name is None):
                self.chroms[chrom].insert( start, end, MyInterval(start,end) )
            else:
                self.chroms[chrom].insert( start, end, MyInterval(start,end,[gene_id, gene_name]) )

    def load_file(self, file):
        if file.endswith('.bed.gz'):
            bed_file = gzip.open(file,'rt')
        elif file.endswith('.bed'):
            bed_file = open(file,'r')
        else:
            raise Exception("Unsupported file type, only .bed or .bed.gz")
        for line in bed_file:
            line = line.rstrip('\n').split('\t')
            if len(line)==3:
                self.insert(line[0], int(line[1]), int(line[2]))
            else:
                self.insert(line[0], int(line[1]), int(line[2]), line[3], line[4])
        bed_file.close()

    def overlaps(self, chrom, start, end):
        if not chrom in self.chroms:
            return False
        overlapping = self.chroms[chrom].find(start, end)
        return len(overlapping)>0

    def closest(self,chrom, start, end):
        if not chrom in self.chroms:
            return ['NA', 'NA', 'NA']

        #first checking if this site overlaps any from the loaded file
        overlapping = self.chroms[chrom].find(start, end)
        if len(overlapping)>0:
            return [overlapping[0].value[0],overlapping[0].value[1],0]

        #to the left
        #it finds features with a start > than 'position'
        left = self.chroms[chrom].before(start, max_dist=1e5)

        #to the right
        #it finds features with a end < than 'position'
        right = self.chroms[chrom].after(end, max_dist=1e5)

        if len(left)>0:
            if len(right)>0:
                dist_left = max(0,1 + start - left[0].end)
                dist_right = max(0,1 + right[0].start - end)
                if dist_left < dist_right:
                    return [left[0].value[0],left[0].value[1],dist_left]
                return [right[0].value[0],right[0].value[1],dist_right]
            else:
                dist_left = max(0,1 + start - left[0].end)
                return [left[0].value[0],left[0].value[1],dist_left]
        else:
            if len(right)>0:
                dist_right = max(0,1 + right[0].start - end)
                return [right[0].value[0],right[0].value[1],dist_right]
            return ['NA', 'NA', 'NA']



'''
Created on May 5, 2021

@author: Juan L. Mateo
'''

from aceofbases.sequence_methods import reverse_complement


class EditingSite:
    "A class to define the targeted editing site"
    def __init__(self, target_seq, sequence, position, strand, fwd_primer, rev_primer, max_ot = float('inf')):
        self.sequence = sequence
        self.position = position  # leftmost coordinates, including the PAM if 5'
        self.strand = strand
        self.score = 100
        self.label = None
        self.oligo1 = ""  # leading GG, forward
        self.oligo2 = ""  # leading GG, reverse
        self.oligo_a_fwd = ""  # adding, forward
        self.oligo_a_rev = ""  # adding, reverse
        self.oligo_s_fwd = ""  # substituting, forward
        self.oligo_s_rev = ""  # substituting, reverse
        self.off_targets = []
        self.max_ot = max_ot

        # oligos
        if fwd_primer == "TAGG":  # T7
            if target_seq[0] == 'G' and target_seq[1] == 'G':
                self.oligo1 = 'TA' + target_seq
                self.oligo2 = 'AAAC' + reverse_complement(target_seq[2:])
            elif target_seq[0] == 'G' and not target_seq[1] == 'G':
                self.oligo_a_fwd = 'TAg' + target_seq
                self.oligo_a_rev = 'AAAC' + reverse_complement(self.oligo_a_fwd[4:])
                self.oligo_s_fwd = 'TAGg' + target_seq[2:]
                self.oligo_s_rev = 'AAAC' + reverse_complement(self.oligo_s_fwd[4:])
            else:
                self.oligo_a_fwd = 'TAgg' + target_seq
                self.oligo_a_rev = 'AAAC' + reverse_complement(self.oligo_a_fwd[4:])
                self.oligo_s_fwd = 'TAgg' + target_seq[2:]
                self.oligo_s_rev = 'AAAC' + reverse_complement(self.oligo_s_fwd[4:])
        elif fwd_primer == "CACCG":
            if target_seq[0] == 'G':
                self.oligo1 = 'CACC' + target_seq
                self.oligo2 = 'AAAC' + reverse_complement(target_seq)
            else:
                self.oligo_a_fwd = 'CACCg' + target_seq
                self.oligo_a_rev = 'AAAC' + reverse_complement('G' + self.oligo_a_fwd[5:])
                self.oligo_s_fwd = 'CACCg' + target_seq[1:]
                self.oligo_s_rev = 'AAAC' + reverse_complement('G' + self.oligo_s_fwd[5:])
        else:
            self.oligo1 = fwd_primer + target_seq
            self.oligo2 = rev_primer + reverse_complement(target_seq)

    def add_offtarget(self, offtarget, exons, genes):
        "Adds a new off-target to this site"
        if len(self.off_targets)>self.max_ot:
            return

        offtarget.set_gene_info(exons, genes)
        self.off_targets.append(offtarget)


class EditingSites:
    "A class as a container of editing sites"
    def __init__(self):
        self.sites = dict() # sites indexes by the label

    def add(self, target_seq, sequence, position, strand, fwd_primer, rev_primer, base_editor, max_ot = float('inf')):
        "Adds a new editing site to the container"
        new_site = EditingSite(target_seq, sequence, position, strand, fwd_primer, rev_primer, max_ot)
        # initial label
        new_site.label = "C" + str(len(self.sites)+1)
        new_site.score = base_editor.score_target(sequence)
        self.sites[new_site.label] = new_site

    def get_sites(self):
        "Returns the editing sites as a list"
        return list(self.sites.values())
    def get_sites_sorted(self):
        "Returns the editing sites sorted by label"
        # notice that to sort the keys we discard the leading character ('T')
        return [value for (key, value) in sorted(list(self.sites.items()),key= lambda item: int(item[0][1:]))]

    def get_site(self, label):
        "Returns the site identified by its label"
        return self.sites[label]

    def sort_offtargets(self):
        "Sorts the off-targets based on their score"
        for site in list(self.sites.values()):
            site.off_targets.sort(key=lambda offtarget: (offtarget.score,
                                                        (lambda dist: 0 if dist=="NA" else dist)(offtarget.distance)))

    def scale_score_and_relabel(self):
        "Sorts the sites by score and relabel them starting by 'T1'"
        new_dict = dict()
        sorted_sites_by_score = list(self.sites.values())
        sorted_sites_by_score.sort(key=lambda site: (site.score), reverse=True)

        for idx, site in enumerate(sorted_sites_by_score):
            new_label = 'T' + str(idx + 1)
            site.label = new_label
            new_dict[new_label] = site

        self.sites = new_dict




#GFF3 format definition
#https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md

import argparse
import gzip
import sys

class Gene:
    def __init__(self, chrom, start, end, strand, geneId, geneName=''):
        self.chrom = chrom
        self.strand = strand
        self.geneId = geneId
        self.geneName = geneName
        self.start = start
        self.end = end
        self.exons = set()
        
    def setGeneName(self, name):
        self.geneName = name
    def setGeneId(self, geneId):
        self.geneId = geneId
    def addExon(self, start, end):
        self.exons.add(Exon(start, end))
        
    def toGeneFile(self):
        return '%s\t%s\t%s\n' % (self.chrom, self.start - 1, self.end)
        
    def toExonFile(self):
        if len(self.exons) == 0:
            # transcript without exons
            return '%s\t%s\t%s\t%s\t%s\n' % (self.chrom, self.start - 1, self.end, self.geneId, self.geneName)
        string = ''
        for exon in list(self.exons):
            string = string + '%s\t%s\t%s\t%s\t%s\n' % (self.chrom, exon.start - 1, exon.end, self.geneId, self.geneName)
        return string
    
    def __eq__(self, other):
        return self.chrom == other.chrom and self.strand == other.strand and self.exons == other.exons
    
class Exon:
    def __init__(self, start, end):
        self.start = start
        self.end = end
    def __eq__(self, other):
        return self.start == other.start and self.end == other.end
    def __hash__(self):
        return hash((self.start, self.end))
        
class Genes:
    def __init__(self):
        self.genes = dict()
        
    def add(self, gene_id, gene):
        return self.genes.setdefault(gene_id,gene)
        
    def getGeneContent(self):
        return [x for x in (gene.toGeneFile() for gene in list(self.genes.values()))]
    
    def getExonContent(self):
        allExons = ''
        for gene in list(self.genes.values()):
            allExons = allExons + gene.toExonFile()
        return allExons

def findTopParent(parent,parents, parentSoFar = None):
    if not parent in parents:
        return parentSoFar
    return findTopParent(parents[parent][0], parents, parents[parent])
def readGFF(gff):
    genes = Genes()
    parents = {}
    if (gff.readline().strip()!='##gff-version 3'):
        raise Exception("The input file is not in gff version 3 format")
    for line in gff:
        if (line.startswith('#')):
            continue
        line = line.strip()
        if (len(line) == 0):
            continue
        
        cols = line.split('\t')
        if (len(cols) != 9):
            continue
        
        seqname = cols[0]
        feature = cols[2]
        start = int(cols[3])
        end = int(cols[4])
        strand = cols[6]
        attributes = cols[8]
        f_id = None
        f_name = ''
        f_parent = None
        
        if feature != 'exon':
            idx1 = attributes.find('ID')
            if(idx1==-1):
                continue
            idx2 = attributes.find(';',idx1)
            if(idx2==-1): # if it is the last attribute, go until the end
                idx2 = len(attributes)
            f_id = attributes[idx1+3:idx2]
            idx1 = attributes.find('Parent')
            if(idx1!=-1):
                idx2 = attributes.find(';',idx1)
                if(idx2==-1): # if it is the last attribute, go until the end
                    idx2 = len(attributes)
                f_parent = attributes[idx1+7:idx2]
            idx1 = attributes.find('Name')
            if(idx1!=-1):
                idx2 = attributes.find(';',idx1)
                if(idx2==-1): # if it is the last attribute, go until the end
                    idx2 = len(attributes)
                f_name = attributes[idx1+5:idx2]
            
            parents.setdefault(f_id,[f_parent,seqname,start,end,strand,f_id,f_name])
            if feature == 'gene':
                genes.add(f_id,Gene(seqname,start,end,strand,f_id,f_name))
            continue
            
        if feature == 'exon':
            idx1 = attributes.find('Parent')
            idx2 = attributes.find(';',idx1)
            if(idx2==-1): # if it is the last attribute, go until the end
                idx2 = len(attributes)
            exon_parent = attributes[idx1+7:idx2]
            
            parent_info = findTopParent(exon_parent, parents)
            if parent_info is None:
                sys.stderr.write('Parent %s not found for exon\n' %exon_parent )
            else:
                try:
                    gene = genes.genes[parent_info[5]]
                except KeyError:
                    gene = genes.add(parent_info[5],Gene(parent_info[1],parent_info[2],parent_info[3],parent_info[4],parent_info[5],parent_info[6]))
                gene.addExon(start, end)
                    
            
    return(genes)

def create_files(input_filename, prefix, folder):
    if input_filename.endswith("gz"):
        with gzip.open(input_filename, 'rt') as gff:
            genes = readGFF(gff)
    else:
        with open(input_filename, 'rt') as gff:
            genes = readGFF(gff)
    
    with gzip.open(folder + prefix + '_genes.bed.gz','wt') as geneFile:
       	for line in genes.getGeneContent():
       		geneFile.write(line)
    with gzip.open(folder + prefix +'_exons.bed.gz','wt') as exonFile:
       	exonFile.write(genes.getExonContent())

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help="the input gff file (gzipped)")
    parser.add_argument("prefix", help="the prefix used for the name of the output files")
    parser.add_argument('--folder', default = './',help='folder where to put the output files, default current folder')
    args = parser.parse_args()
    
    create_files(args.input, args.prefix, args.folder)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

'''
Created on 16 Jun 2015

@author: juanlmateo
'''

import sys
import re
from argparse import ArgumentTypeError, ArgumentParser, RawTextHelpFormatter, FileType
import textwrap
from subprocess import Popen, PIPE, DEVNULL
import os
import math

from bx.seq.twobit import TwoBitFile

from aceofbases.bed_interval import BedInterval
from aceofbases.generate_target_site_svg import get_svg
from aceofbases.sequence_methods import build_expression, reverse_complement, iupac_code
from aceofbases.sequence_methods import CDS, get_translations, get_window_translations
from aceofbases.editing_sites import EditingSites
from aceofbases.base_editor import factory
from aceofbases.off_target import OffTarget


OPTIMAL_WINDOW_BG = "rgba(153, 204, 255, 1)"
EXTENDED_WINDOW_BG = "rgba(153, 204, 255, 0.5)"


def run_bowtie(options, bowtie_path, index_path, input_params, output_params):
    '''
    Runs the bowtie command to find hits in the genome
    It expects a list of options parameters,
    the path to the bowtie binary,
    the path to the index incluiding the index name,
    a list with the input parameters
    and another list with the output parameters
    or None if the function should return the output from bowtie
    '''
    bowtie_output = []# list of lines

    cmd = [bowtie_path + 'bowtie'] + options + [index_path] + input_params
    if output_params is None:# the output from bowtie must be returned as a PIPE
        p = Popen(cmd, stdout=PIPE, stderr=DEVNULL, universal_newlines=True)
        if p.wait() == 0:
            bowtie_output = bowtie_output + p.stdout.readlines()
    else:
        cmd = cmd + output_params
        p = Popen(cmd, stderr=DEVNULL)
    if p.wait() != 0:
        sys.stderr.write("Error running Bowtie (" + str(p.wait()) + ")")
        sys.stderr.write(";".join(cmd))
        raise RuntimeError
    if output_params is None:
        return bowtie_output
    return None

def get_formatted_coords(coords):
    return coords[0] + ":" + str(coords[1]) + "-" + str(coords[2])

def get_wrapped_seq(seq, length):
    if len(seq) <= length:
        return seq
    ret_seq = ''
    for idx in range(int(math.ceil(len(seq) / float(length)))):
        ret_seq = ret_seq + seq[length * idx:length * idx + length] + "<br>"
    return ret_seq[:-4]

def show_u6_termination_warning(seq, fwd_primer):
    if(fwd_primer == 'CACCG' and seq.find('TTTT')!=-1):
        return '<span class="hover"><img class="moveimage" src="../../problem_icon.png" style="width: 11px; height: 11px">\
                    <div style="font-family:sans-serif;" class=tooltip>You selected to use the U6 promoter and this cantidate contains a possible termination signal for RNA Pol III.</div> </span>'
    return ''

def get_formatted_ot_position(distance, intragenic):
    if distance == 0:
        return "</td><td bgcolor='ff9999' align='middle'>E"
    if intragenic:
        return "</td><td bgcolor='ffff99' align='middle'>I"
    return "</td><td bgcolor='99ff99' align='middle'>-"

def get_plain_ot_position(distance, intragenic):
    if distance == 0:
        return "Exonic"
    if intragenic:
        return "Intronic"
    return "Intergenic"

def get_formatted_target_seq(candidate, base_editor):

    html = '%s<span style="background:' + EXTENDED_WINDOW_BG + '">%s</span>\
<span style="background:' + OPTIMAL_WINDOW_BG + '">%s</span>\
<span style="background:' + EXTENDED_WINDOW_BG + '">%s</span>\
%s</td>\
    <td>&nbsp;&nbsp;Window translation: <span name="candidateFrame"></span></td><td class="mono"><span name="candidateTranslation" id="%s"></span>'
    if base_editor.get_pam().is5prime:
        pam_length = len(base_editor.get_pam().pam_str)
        return html % (candidate.sequence[0:base_editor.get_window_start()+pam_length],
                   candidate.sequence[base_editor.get_window_start()+pam_length:base_editor.get_optimal_window_start()+pam_length],
                   candidate.sequence[base_editor.get_optimal_window_start()+pam_length:base_editor.get_optimal_window_end()+pam_length],
                   candidate.sequence[base_editor.get_optimal_window_end()+pam_length:base_editor.get_window_end()+pam_length],
                   candidate.sequence[base_editor.get_window_end()+pam_length:],
                       candidate.label)
    return html % (candidate.sequence[0:base_editor.get_window_start()],
                   candidate.sequence[base_editor.get_window_start():base_editor.get_optimal_window_start()],
                   candidate.sequence[base_editor.get_optimal_window_start():base_editor.get_optimal_window_end()],
                   candidate.sequence[base_editor.get_optimal_window_end():base_editor.get_window_end()],
                   candidate.sequence[base_editor.get_window_end():],
                   candidate.label)


def get_formatted_ot_seq(seq, core_length, mm, base_editor):
    pam_size = len(base_editor.get_pam().pam_str)
    if mm.startswith("PAM"):  # the PAM is 5'
        if core_length != "NA" and core_length > 0:
            tmp_seq = 'PAM[' + seq[pam_size:pam_size + core_length] + ']' + seq[pam_size + core_length:]
        else:
            tmp_seq = 'PAM' + seq[pam_size:]
        ret_seq = ''
        for idx in range(len(mm)):
            if mm[idx] == '-':
                ret_seq = ret_seq + '<span style="font-weight:bold;color:red">' + tmp_seq[idx] + '</span>'
            else:
                ret_seq = ret_seq + tmp_seq[idx]

        return ret_seq[3:]  # it will be displayed without the characters 'PAM'
    # the PAM is 3'
    if core_length != "NA" and core_length > 0:
        tmp_seq = seq[0:base_editor.get_target_size() - (core_length)] + '[' + seq[base_editor.get_target_size() - core_length:-pam_size] + ']' + "PAM"
    else:
        tmp_seq = seq[0:-pam_size] + "PAM"
    ret_seq = ''
    brackets_seen = 0
    for idx in range(len(mm)):
        if mm[idx] in "[]":
            brackets_seen += 1

        if idx==base_editor.get_window_start() + brackets_seen:
            ret_seq = ret_seq + '<span style="background:' + EXTENDED_WINDOW_BG + '">'
        if idx==base_editor.get_optimal_window_start() + brackets_seen:
            ret_seq = ret_seq + '</span><span style="background:' + OPTIMAL_WINDOW_BG + '">'
        if idx==base_editor.get_optimal_window_end() + brackets_seen:
            ret_seq = ret_seq + '</span><span style="background:' + EXTENDED_WINDOW_BG + '">'
        if idx==base_editor.get_window_end() + brackets_seen:
            ret_seq = ret_seq + '</span>'
        if mm[idx] == '-':
            ret_seq = ret_seq + '<span style="font-weight:bold;color:red">' + tmp_seq[idx] + '</span>'
        else:
            ret_seq = ret_seq + tmp_seq[idx]

    return ret_seq[:-3]  # it will be displayed without the characters 'PAM'

def is_within_exons(target_start,strand,target_seq, cds, base_editor):

    if strand=='+':
        pos = target_start + cds.start
        for exon in cds.exons:
            if (exon.start<=pos+base_editor.window_end and pos+base_editor.window_start<=exon.end):
                # now checking if there is editable base
                w_start = max(base_editor.get_window_start(), exon.start - pos - 1)
                w_end = min(base_editor.get_window_end(), exon.end-pos-1)
                return base_editor.get_base_to_edit() in target_seq[w_start:w_end]
        return False
    #reverse
    pos = target_start + len(base_editor.get_pam().pam_str) + len(target_seq)-1 + cds.start
    for exon in cds.exons:
        if (exon.start<=pos-base_editor.window_start and pos-base_editor.window_end<=exon.end):
            w_start = max(base_editor.get_window_start(), pos-exon.end+1)
            w_end = min(base_editor.get_window_end(), pos - exon.start+1)
            return base_editor.get_base_to_edit() in target_seq[w_start:w_end]

    return False

def add_candidate_targets(base_editor, five_prime, three_prime, genomic, cds, strand, candidates, fwd_primer, rev_primer, max_ot = float('inf')):
    reg_exp = build_expression(base_editor.get_pam().pam_str)
    five_prime_re = '^' + build_expression(five_prime)
    three_prime_re = build_expression(three_prime) + '$'
    if base_editor.get_pam().is5prime:
        indices = [m.start() for m in re.finditer('(?=' + reg_exp + ')', genomic, re.I)]
        for index in indices:
            if (index + base_editor.get_target_size() + len(base_editor.get_pam().pam_str)) > len(genomic):
                continue
            candidate_sequence = genomic[index + len(base_editor.get_pam().pam_str):index + len(base_editor.get_pam().pam_str) + base_editor.get_target_size()]
            pam_sequence = genomic[index:index + len(base_editor.get_pam().pam_str)]
            if (not re.search(five_prime_re, candidate_sequence) is None) and (not re.search(three_prime_re, candidate_sequence) is None):
                # we need to transform the index from the reversed sequence to the forward sequence
                if strand == '+':
                    candidates.add(candidate_sequence, pam_sequence + candidate_sequence, index, strand, fwd_primer, rev_primer, base_editor, max_ot)
                else:
                    candidates.add(candidate_sequence, pam_sequence + candidate_sequence, len(genomic) - (index + base_editor.get_target_size() + len(base_editor.get_pam().pam_str)), strand, fwd_primer, rev_primer, base_editor, max_ot)
    else:
        indices = [m.start() for m in re.finditer('(?=' + reg_exp + ')', genomic, re.I)]
        for index in indices:
            if (index - base_editor.get_target_size()) < 0:
                continue
            candidate_sequence = genomic[index - base_editor.get_target_size():index]
            pam_sequence = genomic[index:index + len(base_editor.get_pam().pam_str)]
            if (not re.search(five_prime_re, candidate_sequence) is None) and (not re.search(three_prime_re, candidate_sequence) is None):
                #if base_editor.get_base_to_edit() not in candidate_sequence[base_editor.get_window_start():base_editor.get_window_end()]:
                #    continue
                # we need to transform the index from the reversed sequence to the forward sequence
                if strand == '+':
                    # Checking if the candidate is within the exons defined in the cds
                    if is_within_exons(index - base_editor.get_target_size(), strand, candidate_sequence, cds, base_editor):
                        candidates.add(candidate_sequence, candidate_sequence + pam_sequence, index - base_editor.get_target_size(), strand, fwd_primer, rev_primer, base_editor, max_ot)
                else:
                    # Checking if the candidate is within the exons defined in the cds
                    if is_within_exons(len(genomic) - (index + len(base_editor.get_pam().pam_str)), strand, candidate_sequence, cds, base_editor):
                        candidates.add(candidate_sequence, candidate_sequence + pam_sequence, len(genomic) - (index + len(base_editor.get_pam().pam_str)), strand, fwd_primer, rev_primer, base_editor, max_ot)

def find_ot(candidates, base_editor, output_path, core_mismatches, core_range, total_mismatches, bowtie_path, index_path, exons, genes):
    seqs = base_editor.get_pam().get_sequences_for_bowtie(candidates.get_sites())
    with open(output_path + '/bowtie_input.fasta','w') as bowtie_input:
        bowtie_input.write(seqs)

    bowtie_options = base_editor.get_pam().get_bowtie_options(core_mismatches, core_range, total_mismatches)
    run_bowtie(bowtie_options, bowtie_path, index_path, ['-f', output_path + '/bowtie_input.fasta'], [output_path + '/bowtie_output'])

    # process output and add off-target info
    with open(output_path + '/bowtie_output', 'r') as bowtie_output:
        for line in bowtie_output:
            columns = line.split('\t')
            if base_editor.get_pam().is_bowtie_hit_consistent(columns):
                #                     forward?                        chromosome  strand      start            substitutions sequence  lengthSeq        lengthPAM                           core_range
                off_target = OffTarget(base_editor.get_pam().is5prime, columns[2], columns[1], int(columns[3]), columns[7], columns[4], len(columns[4]), len(base_editor.get_pam().pam_str), core_range)
                if base_editor.get_pam().is5prime:
                    pass
                else:
                    if base_editor.get_base_to_edit() not in off_target.sequence[base_editor.get_window_start():base_editor.get_window_end()]:
                        continue
                candidates.get_site(columns[0]).add_offtarget(off_target, exons, genes)
    # cleaning
    os.remove(output_path + '/bowtie_input.fasta')
    os.remove(output_path + '/bowtie_output')

def get_html_translations(candidates, cds, base_editor):
    translations = get_translations(cds.sequence)
    html = '<table>\
<tbody>\
<tr><td>Translation<br>Frame \
<select id="frame"\
    onchange="updateTranslation();">\
    <option value="F1">F1</option>\
    <option value="F2">F2</option>\
    <option value="F3">F3</option>\
    <option value="F-1">F-1</option>\
    <option value="F-2">F-2</option>\
    <option value="F-3">F-3</option>\
</select>\
</td>\
<td class="mono">\
<span id="translation">%s</span>\
</td>\
</tr></tbody>\
</table>\
<script>\
var translationsText = \'%s\';\
var translationsJson = JSON.parse(translationsText);\
function updateTranslation() {\
    var frame = document.getElementById("frame").value;\
    var x = document.getElementsByName("candidateFrame");\
    for (var i =0; i<x.length; i++){\
        x[i].innerHTML = frame;\
    }\
    var x = document.getElementsByName("candidateTranslation");\
    for (var i =0; i<x.length; i++){\
        var label = x[i].id;\
        x[i].innerHTML = translationsJson[label][frame];\
    }\
    var boxes = document.getElementsByClassName("non-synonymous");\
    for (var i =0; i<boxes.length; i++){\
        boxes[i].style.setProperty("display","none");\
    }\
    var boxes = document.getElementsByClassName("nss-" + frame);\
    for (var i =0; i<boxes.length; i++){\
    boxes[i].style.setProperty("display","inline");\
    }\
    if(frame=="F1") document.getElementById("translation").innerHTML="%s";\
    if(frame=="F2") document.getElementById("translation").innerHTML="%s";\
    if(frame=="F3") document.getElementById("translation").innerHTML="%s";\
    if(frame=="F-1") document.getElementById("translation").innerHTML="%s";\
    if(frame=="F-2") document.getElementById("translation").innerHTML="%s";\
    if(frame=="F-3") document.getElementById("translation").innerHTML="%s";\
}\
    window.onload = updateTranslation\
</script>'

    return html % (get_wrapped_seq(translations[0], 60),\
                   get_candidates_json_translations(candidates, cds, base_editor),\
                   get_wrapped_seq(translations[0], 60),get_wrapped_seq(translations[1], 60),get_wrapped_seq(translations[2], 60),\
                   get_wrapped_seq(translations[3], 60),get_wrapped_seq(translations[4], 60),get_wrapped_seq(translations[5], 60))

def get_formatted_translation(translation):
    html = '<table><tr>'
    for idx in range(len(translation.translation)):
        aa = translation.translation[idx]
        if ">" in aa:
            if "*" in aa:
                color = "black;color:white"
            else:
                color = "purple;color:white"
        else:
            color = "white"
            aa = '&nbsp;' + aa + '&nbsp;'

        if translation.window_start_aa <= idx < translation.window_end_aa:
            html =html + '<td style=\\\'background-color:' + OPTIMAL_WINDOW_BG +';padding: 3px\\\'><span style=\\\'background-color:' + color + '\\\'>' + aa + '</span></td>'
        else:
            html = html + '<td>' + aa + '</td>'
    return html + '</tr></table>'

def get_formatted_translations(translations):
    formatted_translations = []

    formatted_translations.append(get_formatted_translation(translations.frames[0]))
    formatted_translations.append(get_formatted_translation(translations.frames[1]))
    formatted_translations.append(get_formatted_translation(translations.frames[2]))
    formatted_translations.append(get_formatted_translation(translations.frames[3]))
    formatted_translations.append(get_formatted_translation(translations.frames[4]))
    formatted_translations.append(get_formatted_translation(translations.frames[5]))
    return formatted_translations

def get_candidates_json_translations(candidates, cds, base_editor):
    text = '{'
    for candidate in candidates:
        translations = get_window_translations(candidate, cds, base_editor)
        formatted_translations = get_formatted_translations(translations)
        text = text + '"%s":{"F1":"%s","F2":"%s","F3":"%s","F-1":"%s","F-2":"%s","F-3":"%s"},'\
            % (candidate.label, formatted_translations[0],formatted_translations[1],formatted_translations[2],\
             formatted_translations[3],formatted_translations[4],formatted_translations[5])
    text = text[:-1] + '}'
    return text


def get_cds_context(seq, blat_path, two_bit_file, output_path):
    '''
    This functions returns a CDS object with the coordinates of the query
    sequence in the target species, plus the genomic sequences. Both as a
    tuple
    '''
    output_filename = output_path + "output.psl"
    query_filename = output_path + "query.fa"
    with open(query_filename,"wt") as f:
        f.write(">query\n" + seq)
    options = ['-minIdentity=98', '-noHead']
    cmd = [blat_path + 'blat'] + [two_bit_file] + [query_filename] + options + [output_filename]
    p = Popen(cmd, stdout=DEVNULL,stderr=PIPE)
    p.communicate()

    if p.returncode != 0:
        msg = "Error running Blat (" + str(p.returncode) + ")"
        raise RuntimeError(msg)

    while not os.path.isfile(output_filename):
        pass
    with open(output_filename, "rt") as o:
        match = o.readlines()

    # cleaning
    os.remove(output_filename)
    os.remove(query_filename)

    if len(match)!=1: # if not a single match (possibly with many exons)
        cds = CDS("unknown",1,len(seq),"+",seq)
        cds.add_exon(1,len(seq))
        return (cds,seq)

    columns = match[0].split("\t")

    # blat returns coordinates in bed format, the object CDS needs gff
    cds = CDS(columns[13], 1, int(columns[0]), columns[8],seq)
    genomic_exon_start = columns[20].split(",")
    exon_length = columns[18].split(",")
    exon_start = 1
    if cds.strand == "+":
        for idx in range(len(exon_length)-1): # There is an extra comman in these fields
            cds.add_exon(exon_start,exon_start-1+int(exon_length[idx]))
            exon_start += int(exon_length[idx])
            # adding intron
            if idx < len(exon_length)-2:
                exon_start += int(genomic_exon_start[idx+1])-int(genomic_exon_start[idx])-int(exon_length[idx])
    else:
        for idx in reversed(range(len(exon_length)-1)): # There is an extra comman in these fields
            cds.add_exon(exon_start,exon_start-1+int(exon_length[idx]))
            exon_start += int(exon_length[idx])
            # adding intron
            if idx > 0:
                exon_start += int(genomic_exon_start[idx])-int(genomic_exon_start[idx-1])-int(exon_length[idx-1])

    with open(two_bit_file, "br") as f:
        two_bit_file = TwoBitFile(f)# ATTENTION: two_bit_file uses bed coordinates -> start-1
        genomic = two_bit_file[cds.chrom.encode("utf-8")][int(columns[15]):int(columns[16])]

    if cds.strand=="-":
        genomic = reverse_complement(genomic)

    return (cds,genomic)

def valid_dinucleotide_iupac(string):
    valid_chars = ['A', 'C', 'G', 'T', 'N'] + list(iupac_code.keys())
    string = string.upper()
    if string != ''.join(c for c in string if c in valid_chars) or len(string) != 2:
        msg = "%r is not a valid dinucleotide sequence" % string
        raise ArgumentTypeError(msg)
    return string

def valid_overhang(string):
    valid_chars = ['A', 'C', 'G', 'T', 'N']
    string = string.upper()
    if string != ''.join(c for c in string if c in valid_chars) or len(string) > 5:
        msg = "%r is not a valid overhang sequence (up to 5 nt)" % string
        raise ArgumentTypeError(msg)
    return string

def read_multi_fasta(input_file):
    '''
    This function parses the content of a file, as the input string, checking that it has a multifasta format
    and returns a list for each entry that is itself another list with the name and sequence.
    '''
    valid_chars = '-_.() abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'

    file_content = []
    seq_line = ""
    state = 0 # to represent a finate-state machine
              # 0 -> intial state, waiting for a header
              # 1 -> header read, waiting for sequence line
              # 2 -> at least one sequence read, waiting for more or another header
    for line in input_file:
        line = line.strip()
        if len(line) == 0:
            continue
        if state == 0:
            if len(line) > 1 and line[0] == ">":  # header
                state = 1
                name = ''.join(c for c in line if c in valid_chars)
                file_content.append([name])
            else:
                raise Exception("The first line of the file should contain a valid header")
        elif state == 1:
            if re.fullmatch('[ACGTNacgtn]+', line) is not None: # seq
                seq_line = seq_line + line
                state = 2
            else:
                raise Exception("Expected a valid nucleotide sequence after \"%s\"" % file_content[-1][0])
        elif state == 2:
            if len(line) > 1 and line[0] == ">" and state==2:  # header
                file_content[-1].append(seq_line)
                seq_line = ""
                name = ''.join(c for c in line if c in valid_chars)
                file_content.append([name])
                state = 1
            elif re.fullmatch('[ACGTNacgtn]+', line) is not None: # seq
                seq_line = seq_line + line
                state = 2
            else:
                raise Exception("Invalid line \"%s\"" % line)
    #At the end the state must be 2
    if state==2:
        file_content[-1].append(seq_line)
    else:
        raise Exception("Unexpected end of the file")

    return file_content

def do_search(name, query, base_editor, total_mismatches, core_length, core_mismatches, five_prime, three_prime, fwd_primer, rev_primer, output_path, bowtie_path, index_path, two_bit, blat_path, exons_file, genes_file, max_ot):
    base_editor = factory(base_editor)

    if not base_editor.get_pam().core_allowed:
        core_length = "NA"
        core_mismatches = "NA"

    # exons and genes
    exons = BedInterval()
    genes = BedInterval()
    if exons_file is not None and genes_file is not None:
        exons.load_file(exons_file)
        genes.load_file(genes_file)

    (cds, genomic) = get_cds_context(query, blat_path, two_bit, output_path)

    candidates = EditingSites()

    add_candidate_targets(base_editor, five_prime, three_prime, genomic, cds, '+', candidates, fwd_primer, rev_primer, max_ot)
    add_candidate_targets(base_editor, five_prime, three_prime, reverse_complement(genomic), cds, '-', candidates, fwd_primer, rev_primer, max_ot)

    if len(candidates.sites) < 1:
        print("No target sites were found.")
        sys.exit(0)

    # finding off-target sites
    find_ot(candidates, base_editor, output_path, core_mismatches, core_length, total_mismatches, bowtie_path, index_path, exons, genes)

    candidates.sort_offtargets()
    candidates.scale_score_and_relabel()

    sorted_sites = candidates.get_sites_sorted()

    # reporting

    plot_svg = get_svg(cds, candidates, base_editor)

    output_html = open(output_path + 'result_' + name +'.html', 'w')

    output_html.write("<! DOCTYPE html>\n<h2>Detailed results</h2><br>\n")
    output_html.write('<head><meta http-equiv="Content-Type" content="text/html; charset=UTF-8">\n')
    
    output_html.write('<style type="text/css"> td.mono {font-family: "Courier New", Courier, monospace}\n')
    output_html.write('table.fancyTable, tr.fancyTable, td.fancyTable{border: 1px solid black;border-collapse: collapse;}\n')
    output_html.write('.moveimage {\
    position: relative;\
    top: 1px;\
}\
.tooltip {\
    font-size: 10px;\
    display: none;\
    position: absolute;\
    border: 1px solid #cccccc;\
    padding: 0px 5px;\
    background-color: #f2f2f2;\
    text-align: justify;\
    box-shadow: 5px 5px 8px #AAA;\
    width: 300px;\
    line-height: 10px;\
}\
.hover:hover .tooltip {\
    z-index: 10;\
    text-decoration: none;\
    display: block;\
    position: absolute;\
}</style></head>')
    output = open(output_path + name + '.xls', 'w')
    fasta_output = ""
    output_html.write("<tr><td>Input:</td><td class='mono'>" + get_wrapped_seq(query, 60) + "</td></tr> </table></td>\n")
    output.write("Input:\t" + query + "\n")
    output_html.write("<td> <table> <tr><td>Editor:</td><td class='mono'>" + base_editor.__class__.__name__ + "</td></tr>\n")
    output.write("Editor:\t" + base_editor.__class__.__name__ + "\n")
    output_html.write("<tr><td>Target site 5&#39; limitation:</td><td class='mono'>" + five_prime + "</td></tr>\n")
    output.write("Target site 5' limitation:\t" + five_prime + "\n")
    output_html.write("<tr><td>Target site 3&#39; limitation:</td><td class='mono'>" + three_prime + "</td></tr>\n")
    output.write("Target site 3' limitation:\t" + three_prime + "\n")
    output_html.write("<tr><td>Core length:</td><td>" + str(core_length) + "</td></tr>\n")
    output.write("Core length:\t" + str(core_length) + "\n")
    output_html.write("<tr><td>Core MM:</td><td>" + str(core_mismatches) + "</td></tr>\n")
    output.write("Core MM:\t" + str(core_mismatches) + "\n")
    output_html.write("<tr><td>Total MM:</td><td>" + str(total_mismatches) + "</td></tr>\n")
    output.write("Total MM:\t" + str(total_mismatches) + "\n")

    output.write("\n")
    output_html.write("</table> </td></tr></table>\n")

    output_html.write(get_html_translations(sorted_sites, cds, base_editor))
    output_html.write('<br>' + plot_svg + '<br>\n')

    # legend for the OT position
    output_html.write('Legend for off-target site positon: <span style="background-color:#ff9999">E = exonic</span>; <span style="background-color:#ffff99">I = intronic</span>; <span style="background-color:#99ff99">- = intergenic</span>\n')

    # now the off-targets table
    for idx in range(0, len(sorted_sites)):
        fasta_output = fasta_output + '>' + sorted_sites[idx].label + '\n'
        fasta_output = fasta_output + sorted_sites[idx].sequence + '\n'

        output_html.write('<table><tr><td id="' + sorted_sites[idx].label + '" style="font-weight:bold;font-size:20px">' + sorted_sites[idx].label + '</td><td>out of ' + str(len(sorted_sites)) + '</td></tr></table>\n')

        if idx > 0:
            output_html.write('<a href="#' + sorted_sites[idx - 1].label + '">&lt;Previous</a>\n')
        else:
            output_html.write('&lt;Previous\n')
        if idx < len(sorted_sites) - 1:
            output_html.write('<a href="#' + sorted_sites[idx + 1].label + '">Next&gt;</a>\n')
        else:
            output_html.write('Next&gt;\n')
        output_html.write('<br>')
        output_html.write('<table><tr>')
        output_html.write('<td>Sequence:</td><td class="mono">' + show_u6_termination_warning(sorted_sites[idx].sequence,fwd_primer) + get_formatted_target_seq(sorted_sites[idx], base_editor) + '</td></tr>\n')

        output_html.write('</table><table>')

        output.write(sorted_sites[idx].label + '\t' + sorted_sites[idx].sequence + '\t' + str(int(sorted_sites[idx].score)) + '\n')

        if sorted_sites[idx].oligo1 != '':
            output_html.write('<tr><td>Oligo pair</td><td>fwd:</td><td class="mono">' + str(sorted_sites[idx].oligo1) + '</td>\n')
            output_html.write('<td>rev:</td><td class="mono">' + str(sorted_sites[idx].oligo2) + '</td></tr></table>\n')

            output.write('Oligo fwd\t' + str(sorted_sites[idx].oligo1) + '\n')
            output.write('Oligo rev\t' + str(sorted_sites[idx].oligo2) + '\n')
        else:
            output_html.write('<tr><td>Oligo pair with 5&#39; extension</td><td>fwd:</td><td class="mono">' + str(sorted_sites[idx].oligo_a_fwd) + '</td>\n')
            output_html.write('<td>rev:</td><td class="mono">' + str(sorted_sites[idx].oligo_a_rev) + '</td></tr>\n')
            if sorted_sites[idx].oligo_s_fwd != "" and sorted_sites[idx].oligo_s_rev != "":
                output_html.write('<tr><td>Oligo pair with 5&#39; substitution</td><td>fwd:</td><td class="mono">' + str(sorted_sites[idx].oligo_s_fwd) + '</td>\n')
                output_html.write('<td>rev:</td><td class="mono">' + str(sorted_sites[idx].oligo_s_rev) + '</td></tr>')
            output_html.write('</table>\n')

            output.write('Oligo adding fwd\t' + str(sorted_sites[idx].oligo_a_fwd) + '\n')
            output.write('Oligo adding rev\t' + str(sorted_sites[idx].oligo_a_rev) + '\n')
            if sorted_sites[idx].oligo_s_fwd != "" and sorted_sites[idx].oligo_s_rev != "":
                output.write('Oligo substituting fwd\t' + str(sorted_sites[idx].oligo_s_fwd) + '\n')
                output.write('Oligo substituting rev\t' + str(sorted_sites[idx].oligo_s_rev) + '\n')
        if len(sorted_sites[idx].off_targets) > 20:
            output_html.write("<div>Top 20 offtarget sites out of " + str(len(sorted_sites[idx].off_targets) + 1) + " (including on target; for full list see xls file)</div>")
        if base_editor.get_pam().is5prime:
            output_html.write("<table class='fancyTable' border=1 width='1000px'><tr><th>Coordinates</th><th>strand</th><th>MM</th><th>PAM</th><th>target_seq</th><th colspan='2'>distance</th><th>gene name</th><th>gene id</th></tr>\n")
            output.write('Chromosome\tstart\tend\tstrand\tMM\tPAM\ttarget_seq\talignment\tdistance\tposition\tgene name\tgene id\n')
        else:
            output_html.write("<table class='fancyTable' border=1 width='1000px'><tr><th>Coordinates</th><th>strand</th><th>MM</th><th>target_seq</th><th>PAM</th><th colspan='2'>distance</th><th>gene name</th><th>gene id</th></tr>\n")
            output.write('Chromosome\tstart\tend\tstrand\tMM\ttarget_seq\tPAM\talignment\tdistance\tposition\tgene name\tgene id\n')
        for idx2 in range(0, len(sorted_sites[idx].off_targets)):
            # in html output only top 20 offtarget sites
            off_target = sorted_sites[idx].off_targets[idx2]
            if idx2 < 20:
                output_html.write("<tr><td>" + get_formatted_coords(off_target.get_genomic_coordinates()))

                output_html.write("</td><td align='middle'>" + off_target.strand + "</td><td align='middle'>" + str(off_target.mismatches))
                if base_editor.get_pam().is5prime:
                    output_html.write("</td><td class='mono'>" + off_target.sequence[:len(base_editor.get_pam().pam_str)] + "</td><td class='mono'>" + get_formatted_ot_seq(off_target.sequence, core_length, off_target.alignment, base_editor))
                else:  # is the off-target site PAM is NAG the background of the cell is grey
                    output_html.write("</td><td class='mono'>" + get_formatted_ot_seq(off_target.sequence, core_length, off_target.alignment, base_editor) + "</td><td class='mono'>" + off_target.sequence[-len(base_editor.get_pam().pam_str):])
                output_html.write("</td><td>" + str(off_target.distance) + get_formatted_ot_position(off_target.distance, off_target.intragenic))
                output_html.write("</td><td>" + off_target.gene_name + "</td><td>")
                output_html.write(off_target.gene_id + "</td></tr>\n")

            output.write("\t".join(off_target.get_genomic_coordinates()))
            output.write("\t" + off_target.strand)
            output.write("\t" + str(off_target.mismatches))
            if base_editor.get_pam().is5prime:
                output.write("\t" + off_target.sequence[:len(base_editor.get_pam().pam_str)] + "\t" + off_target.sequence[len(base_editor.get_pam().pam_str):])
            else:
                output.write("\t" + off_target.sequence[:-len(base_editor.get_pam().pam_str)] + "\t" + off_target.sequence[-len(base_editor.get_pam().pam_str):])
            output.write("\t" + off_target.alignment + "\t" + str(off_target.distance) + "\t" + get_plain_ot_position(off_target.distance, off_target.intragenic))
            output.write("\t" + off_target.gene_name + "\t" + off_target.gene_id + "\n")
        output_html.write("</table>\n")
        output.write("\n")

    output_html.write("".join("<br>"*15) + "\n")
    output_html.close()
    output.close()
    fasta = open(output_path + name + '.fasta', 'w')
    fasta.write(fasta_output)
    fasta.close()

def main():
    parser = ArgumentParser(formatter_class=RawTextHelpFormatter, description="ACEofBASEs is a careful evaluation of BaseEdits.", epilog=textwrap.dedent('''\
        If you use this tool please cite it as:
        
        
        Have fun using ACEofBASEs!
        '''))
    parser.add_argument("--input", metavar="<file>", type=FileType('r'), help="Fasta file containing the sequence(s) to be scanned for sgRNA candidates.", required=True)
    parser.add_argument("--index", metavar="<file>" , help="Path to the bowtie index files including the name of the index.", required=True)
    parser.add_argument("--bowtie", metavar="<folder>", help="Path to the folder where the executable bowtie is.", default="")
    parser.add_argument("--twobit", metavar="<file>", help="Path to the twoBit file containing the genome sequence of the target species.", required=True)
    parser.add_argument("--blat", metavar="<folder>", help="Path to the folder where the executable blat is.", default="")
    parser.add_argument("--editor", help="Base editor. (default: %(default)s)", default="BE4-Gam", choices=['BE4-Gam', 'ancBE4max', 'evoBE4max', 'ABE8e'])
    parser.add_argument("--five_prime", metavar="<sequence>", type=valid_dinucleotide_iupac, help="Filter candidates target sites with the most 5 prime nucleotides defined by this sequence. IUPAC code allowed. (default: %(default)s)", default="NN")
    parser.add_argument("--three_prime", metavar="<sequence>", type=valid_dinucleotide_iupac, help="Filter candidates target sites with the most 5 prime nucleotides defined by this sequence. IUPAC code allowed. (default: %(default)s)", default="NN")
    parser.add_argument("--fwdOverhang", metavar="<sequence>", type=valid_overhang, help="Sequence of the 5 prime forward cloning oligo. (default: %(default)s)", default="TAGG")
    parser.add_argument("--revOverhang", metavar="<sequence>", type=valid_overhang, help="Sequence of the 5 prime reverse cloning oligo. (default: %(default)s)", default="AAAC")
    parser.add_argument("--totalMM", metavar="<int>", help="Number of total maximum mismatches allowed in the off-target sites. (default: %(default)s)", default=4, type=int)
    parser.add_argument("--core_length", metavar="<int>", help="Number of bases that enclose the core of the target site. (default: %(default)s)", default=12, type=int)
    parser.add_argument("--coreMM", metavar="<int>", help="Number of maximum mismatches allowed in the core of the off-target sites. (default: %(default)s)", default=2, type=int)
    parser.add_argument("--maxOT", metavar="<int>", help="Maximum number of off-target sites to be reported. (default: %(default)s)", default=float("inf"), type=int)
    parser.add_argument("--output", metavar="<folder>", help="Output folder. (default: %(default)s)", default="." + os.path.sep)
    parser.add_argument("--exons_file", metavar="<file>", help="Path to the pseudo-bed file containing the coordinate of exons in the target genome. (default: NotUsed)", default=None)
    parser.add_argument("--genes_file", metavar="<file>", help="Path to the pseudo-bed file containing the coordinate of genes in the target genome. (default: NotUsed)", default=None)
    args = parser.parse_args()

    with args.input:
        file_content = read_multi_fasta(args.input)

    for sequence in file_content:
        sys.stdout.write("Working on sequence '%s'\n" % sequence[0])
        do_search(sequence[0], sequence[1].upper(), args.editor, args.totalMM, args.core_length, args.coreMM, args.five_prime, args.three_prime, args.fwdOverhang, args.revOverhang, args.output, args.bowtie, args.index, args.twobit, args.blat, args.exons_file, args.genes_file, args.maxOT)




'''
Created on Jul 23, 2018

@author: Juan L. Mateo
'''

iupac_code = {'R':'[AG]', 'Y':'[CT]', 'S':'[GC]', 'W':'[AT]', 'K':'[GT]','M':'[AC]',
              'B':'[CGT]', 'D':'[AGT]', 'H':'[ACT]', 'V':'[ACG]', 'N':'[ACGT]'}
dna_codons = {
    # 'M' - START, '*' - STOP
    "GCT": "A", "GCC": "A", "GCA": "A", "GCG": "A",
    "TGT": "C", "TGC": "C",
    "GAT": "D", "GAC": "D",
    "GAA": "E", "GAG": "E",
    "TTT": "F", "TTC": "F",
    "GGT": "G", "GGC": "G", "GGA": "G", "GGG": "G",
    "CAT": "H", "CAC": "H",
    "ATA": "I", "ATT": "I", "ATC": "I",
    "AAA": "K", "AAG": "K",
    "TTA": "L", "TTG": "L", "CTT": "L", "CTC": "L", "CTA": "L", "CTG": "L",
    "ATG": "M",
    "AAT": "N", "AAC": "N",
    "CCT": "P", "CCC": "P", "CCA": "P", "CCG": "P",
    "CAA": "Q", "CAG": "Q",
    "CGT": "R", "CGC": "R", "CGA": "R", "CGG": "R", "AGA": "R", "AGG": "R",
    "TCT": "S", "TCC": "S", "TCA": "S", "TCG": "S", "AGT": "S", "AGC": "S",
    "ACT": "T", "ACC": "T", "ACA": "T", "ACG": "T",
    "GTT": "V", "GTC": "V", "GTA": "V", "GTG": "V",
    "TGG": "W",
    "TAT": "Y", "TAC": "Y",
    "TAA": "*", "TAG": "*", "TGA": "*"
}

def build_expression(seq):
    "Returns a regular expression to model a IUPAC sequence"
    result = ''
    for char in seq:
        if char in iupac_code:
            result = result + iupac_code[char]
        else:
            result = result + char

    return result


def reverse_complement(sequence):
    "Returns the reverse complement of the input sequence"
    rev_comp = []

    for idx in range(len(sequence) - 1, -1, -1):
        if sequence[idx] == 'A':
            rev_comp = rev_comp + ['T']
        elif sequence[idx] == 'C':
            rev_comp = rev_comp + ['G']
        elif sequence[idx] == 'G':
            rev_comp = rev_comp + ['C']
        elif sequence[idx] == 'T':
            rev_comp = rev_comp + ['A']
        else:
            rev_comp = rev_comp + ['N']
    return "".join(rev_comp)

def get_translations(seq):
    "Returns the 6 translation of the input sequence according to the standard code"
    translations = []
    for init in range(0,3):
        translations.append("".join([dna_codons[seq[pos:pos + 3]] for pos in range(init, len(seq) - 2, 3)]))
    seq = reverse_complement(seq)
    for init in range(0,3):
        translations.append("".join([dna_codons[seq[pos:pos + 3]] for pos in range(init, len(seq) - 2, 3)]))
    return translations

class EditingWindowTranslation:
    "A class that represents a translation of an editing window"
    def __init__(self, translation, window_start_aa, window_end_aa, window_start_nt, window_end_nt):
        self.translation = translation
        self.window_start_aa = window_start_aa
        self.window_end_aa = window_end_aa
        self.window_start_nt = window_start_nt
        self.window_end_nt = window_end_nt

class EditingWindowTranslations:
    "A class that represents the translations of an editing window"
    def __init__(self, translations, padding_sequences, base_editor):
        self.frames = []

        for frame in range(0,3):
            window_start_nt = len(padding_sequences[0])-frame
            window_end_nt = window_start_nt + (base_editor.window_end-base_editor.window_start)
            window_start_aa = window_start_nt//3
            window_end_aa = (window_end_nt-1)//3+1#len(translations[frame]) - len(padding_sequences[1])//3
            self.frames.append(EditingWindowTranslation(translations[frame],window_start_aa,window_end_aa,window_start_nt,window_end_nt))
        for frame in range(0,3):
            window_start_nt = len(padding_sequences[1])-frame
            window_end_nt = window_start_nt + (base_editor.window_end-base_editor.window_start)#len(translations[frame])*3 - len(padding_sequences[0])
            window_start_aa = window_start_nt//3
            window_end_aa = (window_end_nt-1)//3+1#len(translations[frame]) - len(padding_sequences[0])//3
            self.frames.append(EditingWindowTranslation(translations[frame+3],window_start_aa,window_end_aa, window_start_nt,window_end_nt))


class CDS:
    "A class that represents the CDS of a transcript"
    def __init__(self, chrom, start, end, strand, sequence):
        self.chrom = chrom
        self.strand = strand
        self.start = start
        self.end = end
        self.exons = []
        self.length = 0
        self.sequence = sequence
    def add_exon(self, start, end):
        "Adds a new exon to the CDS"
        self.exons.append(Exon(start, end))
        self.length = self.length + end-start + 1

class Exon:
    "A class tha represents an Exons"
    def __init__(self, start, end):
        self.start = start
        self.end = end
    def __eq__(self, other):
        return self.start == other.start and self.end == other.end
    def __hash__(self):
        return hash((self.start, self.end))



def get_window_coords_within_cds(candidate,cds, base_editor):
    '''
    Returns the coordinates of the editing window wrt. the CDS.
    This method assumes that the candidate editing window overlaps with a single exon.
    It returns coordinates valid to slice the cds sequence.
    '''
    exon_length = 0
    if candidate.strand=='+':
        pos = candidate.position + cds.start
        for exon in cds.exons:
            if (exon.start<=pos+base_editor.window_end and pos+base_editor.window_start<=exon.end):
                w_start = max(base_editor.get_window_start()+pos-exon.start, 0)
                w_end = min(base_editor.get_window_end()+pos-exon.start, exon.end-exon.start+1)
                return [w_start+exon_length,w_end+exon_length]
            exon_length = exon_length + exon.end - exon.start + 1
        return None
    # reverse
    pos = candidate.position + len(candidate.sequence) - 1 + cds.start
    for exon in cds.exons:
        if (exon.start<=pos-base_editor.window_start and pos-base_editor.window_end<=exon.end):
            w_start = max(pos-base_editor.get_window_end()-exon.start+1, 0)
            w_end = min(pos-base_editor.get_window_start()-exon.start+1, exon.end-exon.start+1)
            return [w_start+exon_length,w_end+exon_length]
        exon_length = exon_length + exon.end - exon.start + 1
    return None


def get_padding_sequences(cds,window_coords):
    "Returns the sequence of the editing window padded with flaking sequences form the CDS"
    pad = 9
    upstream = window_coords[0] - pad - window_coords[0]%3
    upstream = max(upstream,0)
    downstream = window_coords[1] + pad + (len(cds.sequence)-window_coords[1])%3
    downstream = min(downstream,len(cds.sequence))
    return [cds.sequence[upstream:window_coords[0]],cds.sequence[window_coords[1]:downstream]]

baseEditors = {'C':'T','A':'G'}

def get_window_translations(candidate, cds, base_editor):
    "Returns the 6 translations of the editing window"
    window_coords = get_window_coords_within_cds(candidate,cds,base_editor)
    padding_sequences = get_padding_sequences(cds, window_coords)

    window_seq = cds.sequence[window_coords[0]:window_coords[1]]
    if candidate.strand=='+':
        edited_seq = window_seq.replace(base_editor.base_to_edit,baseEditors[base_editor.base_to_edit])
    else:
        temp = reverse_complement(window_seq)
        edited_seq = temp.replace(base_editor.base_to_edit,baseEditors[base_editor.base_to_edit])
        edited_seq = reverse_complement(edited_seq)

    for_translation = padding_sequences[0] + window_seq + padding_sequences[1]
    for_translation_edited = padding_sequences[0] + edited_seq + padding_sequences[1]

    translations = [list(i) for i in get_translations(for_translation)]
    translations_edited = [list(i) for i in get_translations(for_translation_edited)]

    # Annotating the substitutions
    for frame,translation in enumerate(translations):
        for pos,residue in enumerate(translation):
            if residue != translations_edited[frame][pos]:
                translations[frame][pos] =  residue + ">" + translations_edited[frame][pos]

    return EditingWindowTranslations(translations, padding_sequences, base_editor)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="aceofbases",
    version="1.0.0",
    packages=["aceofbases"],
    
    python_requires=">=3.5",
    
    install_requires=["bx-python"],
    
    entry_points = {
        'console_scripts': ['aceofbases=aceofbases.main:main',
        'gff2bedFiles=aceofbases.gff2bedFiles:main'],
    },
    
    author="Juan L. Mateo",
    author_email="mateojuan@uniovi.es",
    description="ACEofBASEs, a careful evaluation of BaseEdits",
    long_description = long_description,
    long_description_content_type="text/markdown",
    keywords="CRISPR",
    url="https://bitbucket.org/juanlmateo/aceofbases_standalone",
    project_urls={
        "Bug Tracker": "https://bitbucket.org/juanlmateo/aceofbases_standalone/issues",
    },
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License"
    ],
)
